from typing import Dict

from fastapi import APIRouter

from .util import call_pipeline
from ..accessors import GenericImageDataAccessor
from ..models import Model
from .params import PipelineParams, PipelineRecord
from ..process import smooth
from ..util import PipelineTrace

from pydantic import Field

router = APIRouter(
    prefix='/pipelines',
    tags=['pipelines'],
)


class SegmentParams(PipelineParams):
    accessor_id: str = Field(description='ID(s) of previously loaded accessor(s) to use as pipeline input')
    model_id: str = Field(description='ID(s) of previously loaded segmentation model(s)')
    channel: int = Field(None, description='Channel to use for segmentation; use all channels if empty.')


class SegmentRecord(PipelineRecord):
    pass


@router.put('/segment')
def segment(p: SegmentParams) -> SegmentRecord:
    """
    Run a semantic segmentation model to compute a binary mask from an input image
    """
    return call_pipeline(segment_pipeline, p)

def segment_pipeline(
        accessors: Dict[str, GenericImageDataAccessor],
        models: Dict[str, Model],
        **k
) -> PipelineTrace:
    input_accessor = accessors.get('accessor')
    model = models.get('model')
    d = PipelineTrace()
    d['input'] = input_accessor
    if ch := k.get('channel') is not None:
        d['mono'] = input_accessor.get_mono(ch)
    d['inference'] = model.label_pixel_class(d.last)
    if sm := k.get('smooth') is not None:
        d['smooth'] = d.last.apply(lambda x: smooth(x, sm))
    d['output'] = d.last
    return d